Imports System
Imports System.IO

' Demo logistic regression using Newton-Raphson
' Note: I left in, but commenteed out, a lot of code in that I used for debugging.

Namespace LogisticRegressionNewtonRaphson
    Friend Class LogisticRegressionNRProgram
        Shared Sub Main(ByVal args() As String)
            Try
                Console.WriteLine(vbCrLf & "Begin Logistic Regression with Newton-Raphson demo")

                Dim trainFile As String = "trainFile.txt" ' will be placed in same directory as executable
                Dim testFile As String = "testFile.txt"

                Console.WriteLine(vbCrLf & "Creating 80 lines of synthetic training data and 20 lines of test data")
                MakeRawDataFile(80, 3, trainFile)
                MakeRawDataFile(20, 4, testFile)

                Console.WriteLine(vbCrLf & "First 5 lines of training data file are: " & vbCrLf)
                DisplayRawData(trainFile, 5)

                Console.WriteLine(vbCrLf & "Loading train and test data from files into memory")

                Dim xTrainMatrix()() As Double = LoadRawDataIntoDesignMatrix(trainFile)
                Console.WriteLine(vbCrLf & "First five rows of x training matrix:")
                Console.WriteLine(MatrixAsString(xTrainMatrix, 5, 1, 5))

                Dim yTrainVector() As Double = LoadRawDataIntoYVector(trainFile)
                Console.WriteLine(vbCrLf & "First five rows of y training vector:")
                Console.WriteLine(VectorAsString(yTrainVector, 5, 1, 3))

                Dim xTestMatrix()() As Double = LoadRawDataIntoDesignMatrix(testFile)
                Dim yTestVector() As Double = LoadRawDataIntoYVector(testFile)

                Console.WriteLine("Setting Newton-Raphson algorithm stop conditions")
                Dim maxIterations As Integer = 25
                Dim epsilon As Double = 0.01 ' stop if all new beta values change less than epsilon (algorithm has converged?)
                Dim jumpFactor As Double = 1000.0 ' stop if any new beta jumps too much (algorithm spinning out of control?)
                Console.WriteLine(vbCrLf & "maxIterations = " & maxIterations & "  epsilon = " & epsilon & "  jumpFactor = " & jumpFactor)

                Console.WriteLine(vbCrLf & "Using Newton-Raphson to find beta parameters that best fit the training data")
                Dim beta() As Double = ComputeBestBeta(xTrainMatrix, yTrainVector, maxIterations, epsilon, jumpFactor) ' computing the beta parameters is synonymous with 'training'
                Console.WriteLine(vbCrLf & "Newton-Raphson complete")
                Console.WriteLine(vbCrLf & "The beta vector is: ")
                Console.WriteLine(VectorAsString(beta, Integer.MaxValue, 4, 10))

                Console.WriteLine("Computing accuracy on test data using the beta values")
                Dim acc As Double = PredictiveAccuracy(xTestMatrix, yTestVector, beta) ' percent of data cases correctly predicted in the test data set.
                Console.WriteLine(vbCrLf & "The predictive accuracy of the model on the test data is " & acc.ToString("F2") & "%" & vbCrLf)

                Console.WriteLine(vbCrLf & "End demo")
                Console.ReadLine()
            Catch ex As Exception
                Console.WriteLine("Fatal: " & ex.Message)
                Console.ReadLine()
            End Try
        End Sub ' Main

        Private Shared Sub MakeRawDataFile(ByVal numLines As Integer, ByVal seed As Integer, ByVal fileName As String)
            Dim ofs As New FileStream(fileName, FileMode.Create)
            Dim sw As New StreamWriter(ofs)

            Dim bValues() As Double = {-95.0, 0.4, -0.9, 11.2} ' hard-coded

            Dim rand As New Random(seed)

            For i = 0 To numLines - 1
                Dim age As Integer = rand.Next(35, 81) ' 35,81) == [35,80
                Dim sex As String = If(rand.NextDouble < 0.5, "M", "F")
                Dim chol As Double = age / 10.0 - 1.0

                Dim flip As Integer = rand.Next(2)
                If flip = 0 Then
                    chol = chol + (3.0 - 1.0) * rand.NextDouble + 1.0
                Else
                    chol = chol - (3.0 - 1.0) * rand.NextDouble + 1.0
                End If
                If chol > 9.9 Then
                    chol = 9.9
                End If
                If chol < 0.1 Then
                    chol = 0.1
                End If

                Dim x0 As Double = 1.0
                Dim x1 As Double = age
                Dim x2 As Double = If(sex = "M", 0.0, 1.0)
                Dim x3 As Double = chol

                Dim z As Double = (bValues(0) * x0) + (bValues(1) * x1) + (bValues(2) * x2) + (bValues(3) * x3)
                Dim p As Double = 1.0 / (1.0 + Math.Exp(-z))

                Dim die As Integer = If(p < 0.5, 0, 1)

                sw.WriteLine(age & " " & sex & " " & chol.ToString("F2") & " " & die)
            Next i
            sw.Close()
            ofs.Close()
        End Sub

        Private Shared Sub DisplayRawData(ByVal fileName As String, ByVal numLines As Integer)
            Dim ifs As New FileStream(fileName, FileMode.Open)
            Dim sr As New StreamReader(ifs)

            Dim line As String = ""
            Dim tokens() As String = Nothing
            Dim ct As Integer = 0

            Console.WriteLine("Age  Sex Chol  |  Died")
            Console.WriteLine("======================")
            line = sr.ReadLine
            Do While line IsNot Nothing AndAlso ct < numLines
                tokens = line.Split(" "c)
                Console.WriteLine(" " & tokens(0) & "   " & tokens(1) & "  " & tokens(2) & "  |   " & tokens(3))
                ct += 1
                line = sr.ReadLine
            Loop
            sr.Close()
            ifs.Close()
            Console.WriteLine(" . . .")
        End Sub

        ' ============================================================================================

        Private Shared Function LoadRawDataIntoDesignMatrix(ByVal rawDataFile As String) As Double()()
            ' read a file of raw age-sex-cholesterol-death (e.g., "68 M 7.56 1") and
            ' add an initial column of all 1.0s to correspond to the B0 constant
            ' the dependent Y variable (death) is loaded by a separate routine

            Dim ifs As New FileStream(rawDataFile, FileMode.Open)
            Dim sr As New StreamReader(ifs)
            Dim line As String = ""
            Dim tokens() As String = Nothing
            Dim ct As Integer = 0
            line = sr.ReadLine
            Do While line IsNot Nothing ' count number lines in file
                ct += 1
                line = sr.ReadLine
            Loop
            sr.Close()
            ifs.Close()

            Dim result()() As Double = MatrixCreate(ct, 4) ' design col, age, sex, cholesterol
            ifs = New FileStream(rawDataFile, FileMode.Open)
            sr = New StreamReader(ifs)
            Dim age As Double
            Dim sex As Double
            Dim chol As Double
            Dim row As Integer = 0

            line = sr.ReadLine
            Do While line IsNot Nothing
                line = line.Trim
                tokens = line.Split(" "c)
                age = CDbl(Integer.Parse(tokens(0)))
                If tokens(1) = "M" Then
                    sex = 0.0
                Else
                    sex = 1.0
                End If
                chol = Double.Parse(tokens(2))

                result(row)(0) = 1.0
                result(row)(1) = age
                result(row)(2) = sex
                result(row)(3) = chol

                row += 1
                line = sr.ReadLine
            Loop
            sr.Close()
            ifs.Close()
            Return result
        End Function ' LoadRawDataIntoDesignMatrix

        Private Shared Function LoadRawDataIntoYVector(ByVal rawDataFile As String) As Double()
            ' read a file of raw age-sex-cholesterol-death (68 M 7.56 1) into a y column vector (1.0 0.0 0.0 1.0 etc)
            Dim ifs As New FileStream(rawDataFile, FileMode.Open)
            Dim sr As New StreamReader(ifs)
            Dim line As String = ""
            Dim tokens() As String = Nothing
            Dim ct As Integer = 0
            line = sr.ReadLine
            Do While line IsNot Nothing ' count number lines in file
                ct += 1
                line = sr.ReadLine
            Loop
            sr.Close()
            ifs.Close()

            Dim result() As Double = VectorCreate(ct) ' single column vector
            ifs = New FileStream(rawDataFile, FileMode.Open)
            sr = New StreamReader(ifs)
            Dim died As Double
            Dim row As Integer = 0

            line = sr.ReadLine
            Do While line IsNot Nothing
                line = line.Trim
                tokens = line.Split(" "c)

                If Integer.Parse(tokens(3)) = 1 Then
                    died = 1.0
                Else
                    died = 0.0
                End If

                result(row) = died

                row += 1
                line = sr.ReadLine
            Loop
            sr.Close()
            ifs.Close()
            Return result
        End Function ' LoadRawDataIntoYVector

        Private Shared Function PredictiveAccuracy(ByVal xMatrix()() As Double, ByVal yVector() As Double, ByVal bVector() As Double) As Double
            ' returns the percent (as 0.00 to 100.00) accuracy of the bVector measured by how many lines of data are correctly predicted.
            ' note: this is not the same as accuracy as measured by sum of squared deviations between 
            ' the probabilities produceed by bVector and 0.0 and 1.0 data in yVector
            ' For predictions we simply see if the p produced by b are >= 0.50 or not.

            Dim xRows As Integer = xMatrix.Length
            Dim xCols As Integer = xMatrix(0).Length
            Dim yRows As Integer = yVector.Length
            Dim bRows As Integer = bVector.Length
            If xCols <> bRows OrElse xRows <> yRows Then
                Throw New Exception("Bad dimensions for xMatrix or yVector or bVector in PredictiveAccuracy()")
            End If

            Dim numberCasesCorrect As Integer = 0
            Dim numberCasesWrong As Integer = 0
            Dim pVector() As Double = ConstructProbVector(xMatrix, bVector) ' helper also used by LogisticRegressionNewtonParameters()
            Dim pRows As Integer = pVector.Length
            If pRows <> xRows Then
                Throw New Exception("Unequal rows in prob vector and design matrix in PredictiveAccuracy()")
            End If

            For i = 0 To yRows - 1 ' each dependent variable
                If pVector(i) >= 0.5 AndAlso yVector(i) = 1.0 Then
                    numberCasesCorrect += 1
                ElseIf pVector(i) < 0.5 AndAlso yVector(i) = 0.0 Then
                    numberCasesCorrect += 1
                Else
                    numberCasesWrong += 1
                End If
            Next i

            Dim total As Integer = numberCasesCorrect + numberCasesWrong
            If total = 0 Then
                Return 0.0
            Else
                Return (100.0 * numberCasesCorrect) / total
            End If
        End Function ' PredictiveAccuracy

        ' ============================================================================================

        Private Shared Function ComputeBestBeta(ByVal xMatrix()() As Double, ByVal yVector() As Double, ByVal maxIterations As Integer, ByVal epsilon As Double, ByVal jumpFactor As Double) As Double()
            ' Use the Newton-Raphson technique to estimate logistic regression beta parameters
            ' xMatrix is a design matrix of predictor variables where the first column is augmented with all 1.0 to represent dummy x values for the b0 constant
            ' yVector is a column vector of binary (0.0 or 1.0) dependent variables
            ' maxIterations is the maximum number of times to iterate in the algorithm. A value of 1000 is reasonable.
            ' epsilon is a closeness parameter: if all new bi values after an iteration are within epsilon of
            ' the old bi values, we assume the algorithm has converged and we return. A value like 0.001 is often reasonable.
            ' jumpFactor stops the algorithm if any new beta value is jumpFactor times greater than the old value. A value of 1000.0 seems reasonable.
            ' The return is a column vector of the beta estimates: b0 is the constant, b1 for x1, etc.
            ' There is a lot that can go wrong here. The algorithm involves finding a matrx inverse (see MatrixInverse) which will throw
            ' if the inverse cannot be computed. The Newton-Raphson algorithm can generate beta values that tend towards infinity. 
            ' If anything bad happens the return is the best beta values known at the time (which could be all 0.0 values but not null).

            Dim xRows As Integer = xMatrix.Length
            Dim xCols As Integer = xMatrix(0).Length

            If xRows <> yVector.Length Then
                Throw New Exception("The xMatrix and yVector are not compatible in LogisticRegressionNewtonParameters()")
            End If

            ' initial beta values
            Dim bVector(xCols - 1) As Double
            For i = 0 To xCols - 1 ' initialize to 0.0. TODO: consider alternatives
                bVector(i) = 0.0
            Next i
            'Console.WriteLine("The initial B vector is")
            'Console.WriteLine(VectorAsString(bVector)) Console.WriteLine(vbCrLf)

            ' best beta values found so far
            Dim bestBvector() As Double = VectorDuplicate(bVector)

            Dim pVector() As Double = ConstructProbVector(xMatrix, bVector) ' a column vector of the probabilities of each row using the bi values and the xi values.
            'Console.WriteLine("The initial Prob vector is: ")
            'Console.WriteLine(VectorAsString(pVector)) Console.WriteLine(vbCrLf)

            'Double)(wMatrix = ConstructWeightMatrix(pVector) ' deprecated. not needed if we use a shortct to comput WX. See ComputeXtilde.
            'Console.WriteLine("The initial Weight matrix is: ")
            'Console.WriteLine(MatrixAsString(wMatrix)); Console.WriteLine(vbCrLf)

            Dim mse As Double = MeanSquaredError(pVector, yVector)
            Dim timesWorse As Integer = 0 ' how many times are the new betas worse (i.e., give worse MSE) than the current betas

            For i = 0 To maxIterations - 1
                'Console.WriteLine("=================================")
                'Console.WriteLine(i)

                Dim newBvector() As Double = ConstructNewBetaVector(bVector, xMatrix, yVector, pVector) ' generate new beta values using Newton-Raphson. could return null.
                If newBvector Is Nothing Then
                    'Console.WriteLine("The ConstructNewBetaVector() helper method in LogisticRegressionNewtonParameters() returned null")
                    'Console.WriteLine("because the MatrixInverse() helper method in ConstructNewBetaVector returned null")
                    'Console.WriteLine("because the current (X'X~) product could not be inverted")
                    'Console.WriteLine("Returning best beta vector found")
                    'Console.ReadLine
                    Return bestBvector
                End If

                'Console.WriteLine("New b vector is ")
                'Console.WriteLine(VectorAsString(newBvector)) Console.WriteLine(vbCrLf)

                ' no significant change?
                If NoChange(bVector, newBvector, epsilon) = True Then ' we are done because of no significant change in beta[]
                    'Console.WriteLine("No significant change between old beta values and new beta values -- stopping")
                    'Console.ReadLine
                    Return bestBvector
                End If
                ' spinning out of control?
                If OutOfControl(bVector, newBvector, jumpFactor) = True Then ' any new beta more than jumpFactor times greater than old?
                    'Console.WriteLine("The new beta vector has at least one value which changed by a factor of " + jumpFactor + " -- stopping")
                    'Console.ReadLine
                    Return bestBvector
                End If

                pVector = ConstructProbVector(xMatrix, newBvector)

                ' are we getting worse or better?
                Dim newMSE As Double = MeanSquaredError(pVector, yVector) ' smaller is better
                If newMSE > mse Then ' new MSE is worse than current SSD
                    timesWorse += 1 ' update counter
                    If timesWorse >= 4 Then
                        'Console.WriteLine("The new beta vector produced worse predictions even after modification four times in a row -- stopping")
                        Return bestBvector
                    End If
                    'Console.WriteLine("The new beta vector has produced probabilities which give worse predictions -- modifying new betas to halfway between old and new")
                    'Console.WriteLine("Times worse = " + timesWorse)

                    bVector = VectorDuplicate(newBvector) ' update current b: old b becomes not the new b but halfway between new and old
                    For k = 0 To bVector.Length - 1
                        bVector(k) = (bVector(k) + newBvector(k)) / 2.0
                    Next k
                    mse = newMSE ' update current SSD (do not update best b because we don't have a new best b)
                    'Console.ReadLine
                Else ' new SSD is be better than old
                    bVector = VectorDuplicate(newBvector) ' update current b: old b becomes new b
                    bestBvector = VectorDuplicate(bVector) ' update best b
                    mse = newMSE ' update current MSE
                    timesWorse = 0 ' reset counter
                End If

                'double pa = PredictiveAccuracy(xMatrix, yVector, bestBvector); // how many cases are we correctly predicting
                'Console.WriteLine("Predictive accuracy is " + pa.ToString("F4"))

                'Console.WriteLine("=================================")
                'Console.ReadLine
            Next i ' end main iteration loop

            'Console.WriteLine("Exceeded max iterations -- stopping")
            'Console.ReadLine
            Return bestBvector
        End Function ' ComputeBestBeta

        ' --------------------------------------------------------------------------------------------

        Private Shared Function ConstructNewBetaVector(ByVal oldBetaVector() As Double, ByVal xMatrix()() As Double, ByVal yVector() As Double, ByVal oldProbVector() As Double) As Double()
            ' this is the heart of the Newton-Raphson technique
            ' bt = bt-1 + inv(X'Wt-1X)X'(y - pt-1)
            '
            ' bt is the new (time t) b column vector
            ' bt-1 is the old (time t-1) vector
            ' X' is the transpose of the X matrix of x data (1.0, age, sex, chol)
            ' Wt-1 is the old weight matrix
            ' y is the column vector of binary dependent variable data
            ' pt-1 is the old column probability vector (computed as 1.0 / (1.0 + exp(-z) where z = b0x0 + b1x1 + . . .)

            ' note: Wt-1 is nxn which could be huge so instead of computing bt = bt-1 + inv(X'Wt-1X)X'(y - pt-1)
            ' compute bt = bt-1 + inv(X'X~)X'(y - pt-1) where X~ is Wt-1X computed directly
            ' the idea is that the vast majority of Wt-1 cells are 0.0 and so can be ignored

            Dim Xt()() As Double = MatrixTranspose(xMatrix) ' X'
            Dim A()() As Double = ComputeXtilde(oldProbVector, xMatrix) ' WX
            Dim B()() As Double = MatrixProduct(Xt, A) ' X'WX

            Dim C()() As Double = MatrixInverse(B) ' inv(X'WX)
            If C Is Nothing Then ' computing the inverse can blow up easily
                Return Nothing
            End If

            Dim D()() As Double = MatrixProduct(C, Xt) ' inv(X'WX)X'
            Dim YP() As Double = VectorSubtraction(yVector, oldProbVector) ' y-p
            Dim E() As Double = MatrixVectorProduct(D, YP) ' inv(X'WX)X'(y-p)
            Dim result() As Double = VectorAddition(oldBetaVector, E) ' b + inv(X'WX)X'(y-p)

            Return result ' could be null!
        End Function ' ConstructNewBvector

        ' --------------------------------------------------------------------------------------------

        Private Shared Function ComputeXtilde(ByVal pVector() As Double, ByVal xMatrix()() As Double) As Double()()
            ' note: Wt-1 is nxn which could be huge so instead of computing bt = bt-1 + inv(X'Wt-1X)X'(y - pt-1) directly
            ' we compute the Wt-1X part, without the use of W.
            ' Since W is derived from the prob vector and W has non-0.0 elements only on the diagonal we can avoid a ton of work
            ' by using the prob vector directly and not computing W at all.
            ' Some of the research papers refer to the product Wt-1X as X~ hence the name of this method.
            ' ex: if xMatrix is 10x4 then W would be 10x10 so WX would be 10x4 -- the same size as X

            Dim pRows As Integer = pVector.Length
            Dim xRows As Integer = xMatrix.Length
            Dim xCols As Integer = xMatrix(0).Length

            If pRows <> xRows Then
                Throw New Exception("The pVector and xMatrix are not compatible in ComputeXtilde")
            End If

            ' we are not doing marix multiplication. the p column vector sort of lays on top of each matrix column.
            Dim result()() As Double = MatrixCreate(pRows, xCols) ' could use (xRows, xCols) here

            For i = 0 To pRows - 1
                For j = 0 To xCols - 1
                    result(i)(j) = pVector(i) * (1.0 - pVector(i)) * xMatrix(i)(j) ' note the p(1-p)
                Next j
            Next i ' i
            Return result
        End Function ' ComputeXtilde

        ' --------------------------------------------------------------------------------------------

        Private Shared Function NoChange(ByVal oldBvector() As Double, ByVal newBvector() As Double, ByVal epsilon As Double) As Boolean
            ' true if all new b values have changed by amount smaller than epsilon
            For i = 0 To oldBvector.Length - 1
                If Math.Abs(oldBvector(i) - newBvector(i)) > epsilon Then ' we have at least one change
                    Return False
                End If
            Next i
            Return True
        End Function ' NoChange

        Private Shared Function OutOfControl(ByVal oldBvector() As Double, ByVal newBvector() As Double, ByVal jumpFactor As Double) As Boolean
            ' true if any new b is jumpFactor times greater than old b
            For i = 0 To oldBvector.Length - 1
                If oldBvector(i) = 0.0 Then ' if old is 0.0 anything goes for the new value
                    Return False
                End If

                If Math.Abs(oldBvector(i) - newBvector(i)) / Math.Abs(oldBvector(i)) > jumpFactor Then ' too big a change.
                    Return True
                End If
            Next i
            Return False
        End Function

        ' --------------------------------------------------------------------------------------------

        Private Shared Function ConstructProbVector(ByVal xMatrix()() As Double, ByVal bVector() As Double) As Double()
            ' p = 1 / (1 + exp(-z) where z = b0x0 + b1x1 + b2x2 + b3x3 + . . .
            ' suppose X is 10 x 4 (cols are: x0 = const. 1.0, x1, x2, x3)
            ' then b would be a 4 x 1 (col vecror)
            ' then result of X times b is (10x4)(4x1) = (10x1) column vector

            Dim xRows As Integer = xMatrix.Length
            Dim xCols As Integer = xMatrix(0).Length
            Dim bRows As Integer = bVector.Length

            If xCols <> bRows Then
                Throw New Exception("xMatrix and bVector are not compatible in ConstructProbVector")
            End If

            Dim result() As Double = VectorCreate(xRows) ' ex: if xMatrix is size 10 x 4 and bVector is 4 x 1 then prob vector is 10 x 1 (one prob for every row of xMatrix)

            Dim z As Double = 0.0
            Dim p As Double = 0.0

            For i = 0 To xRows - 1
                z = 0.0
                For j = 0 To xCols - 1
                    z += xMatrix(i)(j) * bVector(j) ' b0(1.0) + b1x1 + b2x2 +...
                Next j
                p = 1.0 / (1.0 + Math.Exp(-z)) ' consider checking for huge value of Math.Exp(-z) here
                result(i) = p
            Next i
            Return result
        End Function ' ConstructProbVector

        ' --------------------------------------------------------------------------------------------

        Private Shared Function MeanSquaredError(ByVal pVector() As Double, ByVal yVector() As Double) As Double
            ' how good are the predictions? (using an already-calculated prob vector)
            ' note: it is possible that a model with better (lower) MSE than a second model could give worse predictive accuracy.
            Dim pRows As Integer = pVector.Length
            Dim yRows As Integer = yVector.Length
            If pRows <> yRows Then
                Throw New Exception("The prob vector and the y vector are not compatible in MeanSquaredError()")
            End If
            If pRows = 0 Then
                Return 0.0
            End If
            Dim result As Double = 0.0
            For i = 0 To pRows - 1
                result += (pVector(i) - yVector(i)) * (pVector(i) - yVector(i))
                'result += Math.Abs(pVectori - yVectori) 'average absolute deviation approach
            Next i
            Return result / pRows
        End Function

        ' --------------------------------------------------------------------------------------------

        ' ============================================================================================

        Private Shared Function MatrixCreate(ByVal rows As Integer, ByVal cols As Integer) As Double()()
            ' creates a matrix initialized to all 0.0. assume rows and cols > 0
            Dim result(rows - 1)() As Double
            For i = 0 To rows - 1 ' explicit initialization not necessary.
                result(i) = New Double(cols - 1) {}
            Next i
            Return result
        End Function

        Private Shared Function VectorCreate(ByVal rows As Integer) As Double() ' all vectors in Newton-Raphson are single-column vectors
            Dim result(rows - 1) As Double ' we use this technique when we want to make column vector creation explicit
            Return result
        End Function

        Private Shared Function MatrixAsString(ByVal matrix()() As Double, ByVal numRows As Integer, ByVal digits As Integer, ByVal width As Integer) As String
            Dim s As String = ""
            Dim i As Integer = 0
            Do While i < matrix.Length AndAlso i < numRows
                For j = 0 To matrix(i).Length - 1
                    s &= matrix(i)(j).ToString("F" & digits).PadLeft(width) & " "
                Next j
                s &= Environment.NewLine
                i += 1
            Loop
            Return s
        End Function ' MatrixAsString

        Private Shared Function MatrixDuplicate(ByVal matrix()() As Double) As Double()()
            ' allocates/creates a duplicate of a matrix. assumes matrix is not null.
            Dim result()() As Double = MatrixCreate(matrix.Length, matrix(0).Length)
            For i = 0 To matrix.Length - 1 ' copy the values
                For j = 0 To matrix(i).Length - 1
                    result(i)(j) = matrix(i)(j)
                Next j
            Next i
            Return result
        End Function

        Private Shared Function VectorAddition(ByVal vectorA() As Double, ByVal vectorB() As Double) As Double()
            Dim aRows As Integer = vectorA.Length
            Dim bRows As Integer = vectorB.Length
            If aRows <> bRows Then
                Throw New Exception("Non-conformable vectors in VectorAddition")
            End If
            Dim result(aRows - 1) As Double
            For i = 0 To aRows - 1
                result(i) = vectorA(i) + vectorB(i)
            Next i
            Return result
        End Function

        Private Shared Function VectorSubtraction(ByVal vectorA() As Double, ByVal vectorB() As Double) As Double()
            Dim aRows As Integer = vectorA.Length
            Dim bRows As Integer = vectorB.Length
            If aRows <> bRows Then
                Throw New Exception("Non-conformable vectors in VectorSubtraction")
            End If
            Dim result(aRows - 1) As Double
            For i = 0 To aRows - 1
                result(i) = vectorA(i) - vectorB(i)
            Next i
            Return result
        End Function

        Private Shared Function VectorAsString(ByVal vector() As Double, ByVal count As Integer, ByVal digits As Integer, ByVal width As Integer) As String
            Dim s As String = ""
            Dim i As Integer = 0
            Do While i < vector.Length AndAlso i < count
                s &= " " & vector(i).ToString("F" & digits).PadLeft(width) & Environment.NewLine
                i += 1
            Loop
            s &= Environment.NewLine
            Return s
        End Function

        Private Shared Function VectorDuplicate(ByVal vector() As Double) As Double()
            Dim result(vector.Length - 1) As Double
            For i = 0 To vector.Length - 1
                result(i) = vector(i)
            Next i
            Return result
        End Function

        Private Shared Function MatrixTranspose(ByVal matrix()() As Double) As Double()() ' assumes matrix is not null
            Dim rows As Integer = matrix.Length
            Dim cols As Integer = matrix(0).Length ' assume all columns have equal size
            Dim result()() As Double = MatrixCreate(cols, rows) ' note the indexing swap
            For i = 0 To rows - 1
                For j = 0 To cols - 1
                    result(j)(i) = matrix(i)(j)
                Next j
            Next i
            Return result
        End Function ' TransposeMatrix

        Private Shared Function MatrixProduct(ByVal matrixA()() As Double, ByVal matrixB()() As Double) As Double()()
            Dim aRows As Integer = matrixA.Length
            Dim aCols As Integer = matrixA(0).Length
            Dim bRows As Integer = matrixB.Length
            Dim bCols As Integer = matrixB(0).Length
            If aCols <> bRows Then
                Throw New Exception("Non-conformable matrices in MatrixProduct")
            End If

            Dim result()() As Double = MatrixCreate(aRows, bCols)

            For i = 0 To aRows - 1 ' each row of A
                For j = 0 To bCols - 1 ' each col of B
                    For k = 0 To aCols - 1 ' could use k < bRows
                        result(i)(j) += matrixA(i)(k) * matrixB(k)(j)
                    Next k
                Next j
            Next i

            Return result
        End Function ' MatrixProduct

        Private Shared Function MatrixVectorProduct(ByVal matrix()() As Double, ByVal vector() As Double) As Double()
            Dim mRows As Integer = matrix.Length
            Dim mCols As Integer = matrix(0).Length
            Dim vRows As Integer = vector.Length
            If mCols <> vRows Then
                Throw New Exception("Non-conformable matrix and vector in MatrixVectorProduct")
            End If
            Dim result(mRows - 1) As Double ' an n x m matrix times a m x 1 column vector is a n x 1 column vector
            For i = 0 To mRows - 1
                For j = 0 To mCols - 1
                    result(i) += matrix(i)(j) * vector(j)
                Next j
            Next i
            Return result
        End Function

        Private Shared Function MatrixInverse(ByVal matrix()() As Double) As Double()()
            Dim n As Integer = matrix.Length
            Dim result()() As Double = MatrixDuplicate(matrix)

            Dim perm() As Integer
            Dim toggle As Integer
            Dim lum()() As Double = MatrixDecompose(matrix, perm, toggle)
            If lum Is Nothing Then
                Return Nothing
            End If

            Dim b(n - 1) As Double
            For i = 0 To n - 1
                For j = 0 To n - 1
                    If i = perm(j) Then
                        b(j) = 1.0
                    Else
                        b(j) = 0.0
                    End If
                Next j

                Dim x() As Double = HelperSolve(lum, b)

                For j = 0 To n - 1
                    result(j)(i) = x(j)
                Next j
            Next i
            Return result
        End Function

        ' -------------------------------------------------------------------------------------------------------------------

        Private Shared Function HelperSolve(ByVal luMatrix()() As Double, ByVal b() As Double) As Double() ' helper
            ' solve Ax = b if you already have luMatrix from A and b has been permuted
            Dim n As Integer = luMatrix.Length

            ' 1. make a copy of the permuted b vector
            Dim x(n - 1) As Double
            b.CopyTo(x, 0)

            ' 2. solve Ly = b using forward substitution
            For i = 1 To n - 1
                Dim sum As Double = x(i)
                For j = 0 To i - 1
                    sum -= luMatrix(i)(j) * x(j)
                Next j
                x(i) = sum
            Next i

            ' 3. solve Ux = y using backward substitution
            x(n - 1) /= luMatrix(n - 1)(n - 1)
            For i = n - 2 To 0 Step -1
                Dim sum As Double = x(i)
                For j = i + 1 To n - 1
                    sum -= luMatrix(i)(j) * x(j)
                Next j
                x(i) = sum / luMatrix(i)(i)
            Next i

            Return x
        End Function ' HelperSolve

        ' -------------------------------------------------------------------------------------------------------------------

        Private Shared Function MatrixDecompose(ByVal matrix()() As Double, <System.Runtime.InteropServices.Out> ByRef perm() As Integer, <System.Runtime.InteropServices.Out> ByRef tog As Integer) As Double()()
            ' Doolittle's method (1.0s on L diagonal) with partial pivoting
            Dim rows As Integer = matrix.Length
            Dim cols As Integer = matrix(0).Length ' assume all rows have the same number of columns so just use row 0.
            If rows <> cols Then
                Throw New Exception("Attempt to MatrixDecompose a non-square mattrix")
            End If

            Dim n As Integer = rows ' convenience

            Dim result()() As Double = MatrixDuplicate(matrix) ' make a copy of the input matrix

            perm = New Integer(n - 1) {} ' set up row permutation result
            For i = 0 To n - 1
                perm(i) = i
            Next i

            tog = 1 ' toggle tracks number of row swaps. used by MatrixDeterminant

            Dim ajj, aij As Double

            For j = 0 To n - 2 ' each column
                Dim max As Double = Math.Abs(result(j)(j)) ' find largest value in row
                Dim pRow As Integer = j
                For i = j + 1 To n - 1
                    aij = Math.Abs(result(i)(j))
                    If aij > max Then
                        max = aij
                        pRow = i
                    End If
                Next i

                If pRow <> j Then ' if largest value not on pivot, swap rows
                    Dim rowPtr() As Double = result(pRow)
                    result(pRow) = result(j)
                    result(j) = rowPtr

                    Dim tmp As Integer = perm(pRow) ' and swap perm info
                    perm(pRow) = perm(j)
                    perm(j) = tmp

                    tog = -tog ' adjust the row-swap toggle
                End If

                ajj = result(j)(j)
                If Math.Abs(ajj) < 0.00000001 Then ' if diagonal after swap is zero...
                    Return Nothing ' consider a throw
                End If

                For i = j + 1 To n - 1
                    aij = result(i)(j) / ajj
                    result(i)(j) = aij
                    For k = j + 1 To n - 1
                        result(i)(k) -= aij * result(j)(k)
                    Next k
                Next i
            Next j ' main j loop

            Return result
        End Function ' MatrixDecompose

        ' ===========================================================================================
    End Class ' LogisticRegressionNRProgram
End Namespace 'ns
